/*
 *
 *  Copyright (C) 2017, OFFIS e.V.
 *  All rights reserved.  See COPYRIGHT file for details.
 *
 *  This software and supporting documentation were developed by
 *
 *    Open Connections GmbH
 *    Stau 33
 *    D-26122 Oldenburg, Germany
 *
 *
 *  Module:  dcmnet
 *
 *  Author:  Michael Onken
 *
 *  Purpose: Test DcmSCP and DcmSCU classes (only certain aspects so far)
 *
 */


#include "dcmtk/config/osconfig.h"    /* make sure OS specific configuration is included first */

#ifdef WITH_THREADS

#define INCLUDE_CMATH
#include "dcmtk/ofstd/ofstdinc.h"
#include "dcmtk/ofstd/oftest.h"
#include "dcmtk/ofstd/oftimer.h"
#include "dcmtk/dcmnet/scp.h"
#include "dcmtk/dcmnet/scu.h"



static OFLogger t_scuscp_logger= OFLog::getLogger("dcmtk.test.tscuscp");

/** Method that ensures that the current thread is actually sleeping for the
 *  defined number of seconds (at least).
 *  The problem with the regular sleep() function called from OFStandard::sleep
 *  is that it might be interrupted by signals or a network timeout (depending
 *  on the operating system). This methods re-executes OFStandard'S sleep method
 *  until the desired number of seconds have elapsed.
 *  @param sleep The number of seconds to sleep (at least)
 */
static void force_sleep(Uint32 sleep)
{
    OFTimer timer;
    double elapsed = timer.getDiff();
    while (elapsed < (double)sleep)
    {
        // Use ceiling since otherwise we could wait too short
        OFStandard::sleep(OFstatic_cast(unsigned int, ceil(sleep-elapsed)));
        elapsed = timer.getDiff();
    }
}


/** SCP derived from DcmSCP in order to test two types of virtual methods:
 *  <ul>
 *    <li> Notifiers: Called for some events happening in DcmSCP. So far
 *    only the notifiers notifyConnectionTimeout() as and notifyAssociationTermination
 *    are tested</li>
 *    <li> Setters: Functions returning a value in order to enforce
 *    a specific DcmSCP behavior. So far, only stopAfterCurrentAssociation() and
 *    stopAfterConnectionTimeout() are tested.
 *  </ul>
 *  Additionally the SCP derives from OFThread in order to construct the
 *  test case (i.e. send data with SCU at the same time and check results).
 */
struct TestSCP: DcmSCP, OFThread
{

    /** Constructor
     */
    TestSCP() :
        DcmSCP(),
        m_listen_result(EC_NotYetImplemented), // value indicating "not set"
        m_set_stop_after_assoc(OFFalse),
        m_set_stop_after_timeout(OFFalse),
        m_set_reject_calling_host(OFFalse),
        m_stop_after_assoc_result(OFFalse),
        m_stop_after_timeout_result(OFFalse),
        m_notify_connection_timeout_result(OFFalse),
        m_notify_assoc_termination_result(OFFalse)
    {
    }

    /** Clear settings
     */
    void clear()
    {
        m_listen_result = EC_NotYetImplemented;
        m_set_stop_after_assoc = OFFalse;
        m_set_stop_after_timeout = OFFalse;
        m_set_reject_calling_host = OFFalse;
        m_stop_after_assoc_result = OFFalse;
        m_stop_after_timeout_result = OFFalse;
        m_notify_connection_timeout_result = OFFalse;
        m_notify_assoc_termination_result = OFFalse;
    }

    /** Overwrite method from DcmSCP in order to test feature to stop after current
     *  association.
     *  @return Returns value of m_set_stop_after_assoc
     */
    virtual OFBool stopAfterCurrentAssociation()
    {
        if (m_set_stop_after_assoc)
        {
            m_stop_after_assoc_result = OFTrue;
        }
        else
        {
            m_stop_after_assoc_result = OFFalse;
        }
        return m_set_stop_after_assoc;
    }

    /** Overwrite method from DcmSCP in order to test feature to stop after
    *   the SCP's connection timeout occurred in non-blocking mode.
    *   @return Returns value of m_set_stop_after_timeout
    */
    virtual OFBool stopAfterConnectionTimeout()
    {
        if (m_set_stop_after_timeout)
        {
            m_stop_after_timeout_result = OFTrue;
        }
        else
        {
            m_stop_after_timeout_result = OFFalse;
        }
        return m_stop_after_timeout_result;
    }

    /** Overwrite method from DcmSCP in order to test feature that SCP reports
     *  connection timeout in non-blocking mode.
     */
    virtual void notifyConnectionTimeout()
    {
        m_notify_connection_timeout_result = OFTrue;
    }

    /** Overwrite method from DcmSCP in order to test feature that SCP reports
     *  the termination of an association.
     */
    virtual void notifyAssociationTermination()
    {
        m_notify_assoc_termination_result = OFTrue;
    }

    /** Overwrite method from DcmSCP in order to test feature that SCP rejects
     *  a forbidden host.
     *  @param hostOrIP The host name or IP address to be checked
     */
    virtual OFBool checkCallingHostAccepted(const OFString& hostOrIP)
    {
        // In order to test this feature, we do not need to perform actual checking
        (void)hostOrIP;
        if (m_set_reject_calling_host)
        {
            m_calling_host_accepted = OFFalse;
            return OFFalse;
        }
        else
        {
            m_calling_host_accepted = OFTrue;
            return OFTrue;
        }
    }

    /// The result returned my SCP's listen() method
    OFCondition m_listen_result;
    /// If set, the SCP should stop after the currently running association
    OFBool m_set_stop_after_assoc;
    /// If set, the SCP should stop after TCP timeout occurred in non-blocking mode
    OFBool m_set_stop_after_timeout;
    /// If set, the SCP should not allow the calling host (whatever it is)
    OFBool m_set_reject_calling_host;
    /// Indicator whether related virtual function was called and returned accordingly
    OFBool m_stop_after_assoc_result;
    /// Indicator whether related virtual function was called and returned accordingly
    OFBool m_stop_after_timeout_result;
    /// Indicator whether the calling host was accepted
    OFBool m_calling_host_accepted;
    /// Indicator whether related virtual notifier function was called
    OFBool m_notify_connection_timeout_result;
    /// Indicator whether related virtual notifier function was called
    OFBool m_notify_assoc_termination_result;

    /** Method called by OFThread to start SCP operation. Starts listen() loop of DcmSCP.
    */
    virtual void run()
    {
        m_listen_result = listen();
    }

};

// -------------- End of class TestSCP -------------------------


/** Manually configure Verification SOP class on server
 *  @param cfg The SCP configuration to modify
 */
void configure_scp_for_echo(DcmSCPConfig& cfg, T_ASC_SC_ROLE roleOfRequestor = ASC_SC_ROLE_DEFAULT)
{
    cfg.setPort(11112);
    OFList<OFString> xfers;
    xfers.push_back(UID_LittleEndianImplicitTransferSyntax);
    OFCHECK(cfg.addPresentationContext(UID_VerificationSOPClass, xfers, roleOfRequestor).good());
}


/** Send ECHO to SCP
 *  @param called_ae_title The Called AE Title to be used
 *  @param do_release  If OFTrue, SCU should release the association (otherwise
 *         we expect the SCP to do that)
 *  @param secs_after_echo Seconds to wait after sending echo
 */
void scu_sends_echo(
  const OFString& called_ae_title,
  const OFBool expect_assoc_reject = OFFalse,
  const OFBool do_release = OFTrue,
  const int secs_after_echo = 0)
{
    // make sure server is up
    force_sleep(2);
    DcmSCU scu;
    scu.setAETitle("TEST_SCU");
    scu.setPeerAETitle(called_ae_title);
    scu.setPeerHostName("localhost");
    scu.setPeerPort(11112);
    OFList<OFString> xfers;
    xfers.push_back(UID_LittleEndianImplicitTransferSyntax);
    OFCondition result = scu.addPresentationContext(UID_VerificationSOPClass, xfers);
    OFCHECK(result.good());
    result = scu.initNetwork();
    OFCHECK(result.good());
    result = scu.negotiateAssociation();
    if (!expect_assoc_reject)
    {
        OFCHECK(result.good());
        result = scu.sendECHORequest(1);
        OFCHECK(result.good());
        force_sleep(secs_after_echo);
        if (do_release)
        {
            result = scu.releaseAssociation();
            OFCHECK(result.good());
        }
    }
    else
    {
        OFCHECK(result == DUL_ASSOCIATIONREJECTED);
    }
    return;
}


// Test case that checks whether server returns if it is configured to stop
// after current association
OFTEST_FLAGS(dcmnet_scp_stop_after_current_association, EF_Slow)
{
    // Configure SCP for Verification and start it. First we check that the
    // server does *not* return after association without configuring it
    // accordingly
    TestSCP scp;
    DcmSCPConfig& config = scp.getConfig();
    configure_scp_for_echo(config);
    config.setAETitle("STOP_AFTER_ASSOC");
    config.setConnectionBlockingMode(DUL_BLOCK);
    scp.start();

    // Send ECHO, and wait to be sure SCP has time to exit
    scu_sends_echo("STOP_AFTER_ASSOC");
    // Make sure server would have time to return
    force_sleep(2);

    // Check whether all test variables have the correct values
    OFCHECK(scp.m_listen_result == EC_NotYetImplemented); // still listening
    OFCHECK(scp.m_stop_after_assoc_result == OFFalse);
    OFCHECK(scp.m_stop_after_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_connection_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_assoc_termination_result == OFTrue); // SCU released association

    // Tell SCP to stop after association and run SCU again
    scp.clear();
    scp.m_set_stop_after_assoc = OFTrue;
    scu_sends_echo("STOP_AFTER_ASSOC");
    force_sleep(2);

    // Check whether all test variables have the correct values
    OFCHECK(scp.m_listen_result == NET_EC_StopAfterAssociation);
    OFCHECK(scp.m_stop_after_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_connection_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_assoc_termination_result == OFTrue); // SCU released association

    scp.join();
}


// Test case that checks whether server returns with the correct error code if
// configured with an invalid configuration (here: no presentation contexts)
OFTEST_FLAGS(dcmnet_scp_fail_on_invalid_association_configuration, EF_Slow)
{
    // Configure SCP but do not add any presentation contexts
    TestSCP scp;
    DcmSCPConfig& config = scp.getConfig();
    config.setPort(11112);;
    config.setAETitle("TEST_INVALID_CFG");
    config.setConnectionBlockingMode(DUL_NOBLOCK);
    // Ensure that if the server starts listen loop (it should not), the
    // related thread returns anyway.
    config.setConnectionTimeout(3);
    scp.m_set_stop_after_timeout = OFTrue;
    // Start server and check that it returns with the expected return value.
    // In this case we did not configure any presentation contexts, so we
    // expect the related error code.
    scp.start();
    scp.join();
    OFCHECK(scp.m_listen_result == NET_EC_InvalidSCPAssociationProfile);
    OFCHECK(scp.m_stop_after_assoc_result == OFFalse);
    OFCHECK(scp.m_stop_after_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_connection_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_assoc_termination_result == OFFalse);
}


// Test case that checks whether server checks for allowed/disallowed host
// names, i.e. it refuses the association in case the connecting host
// is not accepted by the related virtual method.
OFTEST_FLAGS(dcmnet_scp_fail_on_disallowed_host, EF_Slow)
{
    // Configure SCP for Verification
    TestSCP scp;
    DcmSCPConfig& config = scp.getConfig();
    configure_scp_for_echo(config);
    config.setAETitle("REJECT_HOST");
    config.setConnectionBlockingMode(DUL_BLOCK);
    // Make sure SCP always returns
    scp.m_set_stop_after_assoc = OFTrue;
    // Tell SCP to reject calling host
    scp.m_set_reject_calling_host = OFTrue;
    scp.start();

    scu_sends_echo("REJECT_HOST", OFTrue /* expect that association is being refused */);
    force_sleep(2);  // make sure server would have time to return

    // Check whether all test variables have the correct values
    OFCHECK(scp.m_listen_result == NET_EC_StopAfterAssociation);
    OFCHECK(scp.m_stop_after_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_connection_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_assoc_termination_result == OFTrue);
    // Check whether the calling host has been the reason for
    // refusing the association
    OFCHECK(scp.m_calling_host_accepted == OFFalse);
    scp.join();
}



// Check whether SCP continues after timeout elapsed if timeout handler does
// not tell SCP to stop
OFTEST_FLAGS(dcmnet_scp_no_stop_wo_request_noblock, EF_Slow)
{
    // Configure SCP for Verification, set timeout but do not tell
    // SCP to stop after timeout occurs, and start SCP
    TestSCP scp;
    DcmSCPConfig& config = scp.getConfig();
    configure_scp_for_echo(config);

    config.setAETitle("NO_STOP_NOBLOCK");
    config.setConnectionBlockingMode(DUL_NOBLOCK);
    config.setConnectionTimeout(1);
    scp.m_set_stop_after_timeout = OFFalse;
    scp.start();

    // Check whether all test variables have the correct values, especially that
    // timeout occurred but did not lead to a stop
    force_sleep(5);
    OFCHECK(scp.m_listen_result == EC_NotYetImplemented);      // still listening
    OFCHECK(scp.m_stop_after_assoc_result == OFFalse);
    OFCHECK(scp.m_stop_after_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_connection_timeout_result == OFTrue); // Timeout occurred, but we did not stop
    OFCHECK(scp.m_notify_assoc_termination_result == OFFalse); // SCU released association

    // Stop (already tested in former test cases)
    scp.m_set_stop_after_timeout = OFTrue;
    scp.join();

}


// Check whether SCP does not stop without request in blocking mode
OFTEST_FLAGS(dcmnet_scp_no_stop_wo_request_block, EF_Slow)
{
    // Configure SCP for Verification and to do not ask to exit for any reason
    TestSCP scp;
    DcmSCPConfig& config = scp.getConfig();
    configure_scp_for_echo(config);
    config.setAETitle("NO_STOP_BLOCK");
    config.setConnectionBlockingMode(DUL_BLOCK);
    scp.start();
    scu_sends_echo("NO_STOP_BLOCK");

    // Give some time to SCP so it could return
    force_sleep(3);

    // Check whether all test variables have the correct values
    // SCP did not return yet (check whether it is still connected)
    OFCHECK(scp.m_listen_result == EC_NotYetImplemented); // still listening
    OFCHECK(scp.m_stop_after_assoc_result == OFFalse);
    OFCHECK(scp.m_stop_after_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_connection_timeout_result == OFFalse);
    OFCHECK(scp.m_notify_assoc_termination_result == OFTrue);  // SCU released association

    // Stop (already tested in former test cases), therefore, send another ECHO
    scp.m_set_stop_after_assoc = OFTrue;
    scu_sends_echo("NO_STOP_BLOCK");
    scp.join();
}


// Check association termination notifier does not fire when there was no
// association at all
OFTEST_FLAGS(dcmnet_scp_no_term_notify_without_association, EF_Slow)
{
    // Configure SCP for Verification and tell it stop after 3 seconds.
    TestSCP scp;
    DcmSCPConfig& config = scp.getConfig();
    configure_scp_for_echo(config);
    config.setAETitle("NO_TERM_WO_ASSOC");
    config.setConnectionBlockingMode(DUL_NOBLOCK);
    config.setConnectionTimeout(3);
    scp.m_set_stop_after_timeout = OFTrue;
    scp.start();
    scp.join();
    // Check whether all test variables have the correct values.
    OFCHECK(scp.m_stop_after_assoc_result == OFFalse);
    OFCHECK(scp.m_stop_after_timeout_result == OFTrue);
    OFCHECK(scp.m_notify_connection_timeout_result == OFTrue);
    OFCHECK(scp.m_notify_assoc_termination_result == OFFalse); // no notification
    OFCHECK(scp.m_listen_result == NET_EC_StopAfterConnectionTimeout); // SCP ran into timeout
}


#endif // WITH_THREADS
